use microrm::make_index;
pub use microrm::{Entity, Modelable, Schema};
use serde::{Deserialize, Serialize};

/// Simple key-value store for persistent configuration
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct PersistentConfig {
    pub key: String,
    pub value: String,
}

make_index!(!PersistentConfigIndex, PersistentConfig::Key);

#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Session {
    pub key: String,
    // TODO: add expiry here
}

microrm::make_index!(!SessionKeyIndex, Session::Key);

/// Authentication state for a session. If no challenges are left, it's considered authorized.
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct SessionAuthentication {
    #[microrm_foreign]
    pub realm: RealmID,
    #[microrm_foreign]
    pub session: SessionID,
    #[microrm_foreign]
    pub user: UserID,

    pub challenges_left: Vec<AuthChallengeType>,
}

make_index!(
    !SessionAuthenticationIndex,
    SessionAuthentication::Realm,
    SessionAuthentication::Session,
    SessionAuthentication::User
);

// **** oauth types ****
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Realm {
    pub shortname: String,
}

make_index!(!RealmIndex, Realm::Shortname);

#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Key {
    #[microrm_foreign]
    pub realm: RealmID,
    pub key_id: String,
    #[serde(with = "serde_bytes")]
    pub keydata: Vec<u8>,
}

/// End-user representation object
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct User {
    #[microrm_foreign]
    pub realm: RealmID,
    pub username: String,
}

make_index!(!UserIndex, User::Realm, User::Username);

#[derive(Clone, Copy, Debug, PartialEq, PartialOrd, Modelable, Serialize, Deserialize)]
pub enum AuthChallengeType {
    Username,
    Password,
    TOTP,
    Grid,
    WebAuthn,
}

#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct AuthChallenge {
    #[microrm_foreign]
    pub user: UserID,
    pub challenge_type: AuthChallengeType,
    #[serde(with = "serde_bytes")]
    pub public: Vec<u8>,
    #[serde(with = "serde_bytes")]
    pub secret: Vec<u8>,
    pub enabled: bool,
}

make_index!(AuthChallengeIndex, AuthChallenge::User);

/// User semantic grouping
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Group {
    #[microrm_foreign]
    pub realm: RealmID,
    pub shortname: String,
}

make_index!(!GroupIndex, Group::Realm, Group::Shortname);

/// User membership in group
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct GroupMembership {
    #[microrm_foreign]
    pub group: GroupID,
    #[microrm_foreign]
    pub user: UserID,
}

make_index!(
    !GroupMembershipIndex,
    GroupMembership::Group,
    GroupMembership::User
);

/// OAuth2 client representation
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Client {
    #[microrm_foreign]
    pub realm: RealmID,
    pub shortname: String,
    pub secret: String,
}

make_index!(!ClientNameIndex, Client::Realm, Client::Shortname);

#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct ClientRedirect {
    #[microrm_foreign]
    pub client: ClientID,
    pub redirect: String,
}

make_index!(ClientRedirectIndex, ClientRedirect::Client);

/// Requested group of permissions
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Scope {
    #[microrm_foreign]
    pub realm: RealmID,
    pub shortname: String,
}

make_index!(!ScopeIndex, Scope::Realm, Scope::Shortname);

/// Specific atomic permission
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct Role {
    #[microrm_foreign]
    pub realm: RealmID,
    pub shortname: String,
}

make_index!(!RoleIndex, Role::Realm, Role::Shortname);

/// Role membership in scope
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct ScopeRole {
    #[microrm_foreign]
    pub scope: ScopeID,
    #[microrm_foreign]
    pub role: RoleID,
}

make_index!(!ScopeRoleIndex, ScopeRole::Scope, ScopeRole::Role);

/// Assigned permissions in group
#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct GroupRole {
    #[microrm_foreign]
    pub group: GroupID,
    #[microrm_foreign]
    pub role: RoleID,
}

make_index!(!GroupRoleIndex, GroupRole::Group, GroupRole::Role);

#[derive(Debug, Entity, Serialize, Deserialize)]
pub struct RevokedToken {
    #[microrm_foreign]
    pub user: UserID,
    pub nonce: String,
}

make_index!(!RevokedTokenIndex, RevokedToken::User, RevokedToken::Nonce);

pub fn schema() -> Schema {
    Schema::new()
        // global config types
        .entity::<PersistentConfig>()
        .index::<PersistentConfigIndex>()
        // session types
        .entity::<Session>()
        .index::<SessionKeyIndex>()
        .entity::<SessionAuthentication>()
        .index::<SessionAuthenticationIndex>()
        // oauth types
        .entity::<Realm>()
        .index::<RealmIndex>()
        .entity::<Key>()
        .entity::<User>()
        .index::<UserIndex>()
        .entity::<AuthChallenge>()
        .index::<AuthChallengeIndex>()
        .entity::<Group>()
        .index::<GroupIndex>()
        .entity::<GroupMembership>()
        .index::<GroupMembershipIndex>()
        .entity::<Client>()
        .index::<ClientNameIndex>()
        .entity::<ClientRedirect>()
        .index::<ClientRedirectIndex>()
        .entity::<Scope>()
        .index::<ScopeIndex>()
        .entity::<Role>()
        .index::<RoleIndex>()
        .entity::<ScopeRole>()
        .index::<ScopeRoleIndex>()
        .entity::<GroupRole>()
        .index::<GroupRoleIndex>()
        .entity::<RevokedToken>()
        .index::<RevokedTokenIndex>()
}
