use crate::{config, jwt, schema, UIDCError};
use microrm::{entity::EntityID, prelude::*};

#[derive(Debug)]
pub enum TokenError {
    DBError(microrm::Error),
    InternalError(&'static str),
    RequestError(&'static str),
}

impl From<microrm::Error> for TokenError {
    fn from(value: microrm::Error) -> Self {
        Self::DBError(value)
    }
}

impl From<ring::error::KeyRejected> for TokenError {
    fn from(value: ring::error::KeyRejected) -> Self {
        Self::InternalError("could not load realm key")
    }
}

pub fn generate_auth_token<'a>(
    config: &config::Config,
    qi: &microrm::QueryInterface,
    realm: schema::RealmID,
    client: schema::ClientID,
    user: schema::UserID,
    scopes: impl Iterator<Item = &'a str>,
) -> Result<String, UIDCError> {
    let realm = qi
        .get()
        .by_id(&realm)
        .one()?
        .ok_or(TokenError::RequestError("no such realm"))?;
    let client = qi
        .get()
        .by_id(&client)
        .one()?
        .ok_or(TokenError::RequestError("no such client"))?;
    let user = qi
        .get()
        .by_id(&user)
        .one()?
        .ok_or(TokenError::RequestError("no such user"))?;

    let issuer = format!("{}/{}", config.base_url, realm.shortname,);

    let iat = std::time::SystemTime::now();
    let exp = iat + std::time::Duration::from_secs(config.auth_token_expiry);

    // find all roles the user can possibly have access to
    let mut user_roles = vec![];
    for group in qi
        .get()
        .by(schema::GroupMembership::User, &user.id())
        .all()?
    {
        for group_role in qi.get().by(schema::GroupRole::Group, &group.id()).all()? {
            user_roles.push(group_role.role);
        }
    }

    // find all roles requested by the scopes
    let mut requested_roles = vec![];
    for scope in scopes {
        if let Some(scope) = qi.get().by(schema::Scope::Shortname, scope).one()? {
            for scope_role in qi.get().by(schema::ScopeRole::Scope, &scope.id()).all()? {
                requested_roles.push(scope_role.role);
            }
        }
    }

    user_roles.sort_by_key(|k| k.raw_id());
    user_roles.dedup();
    requested_roles.sort_by_key(|k| k.raw_id());
    requested_roles.dedup();

    // find the intersection between requested roles and the ones the user actually has
    let resulting_roles = requested_roles
        .iter()
        .filter(|req| user_roles.contains(req))
        .map(|role_id| {
            Ok(serde_json::Value::String(
                qi.get()
                    .by_id(role_id)
                    .one()?
                    .ok_or(UIDCError::Abort("inconsistent role state"))?
                    .wrapped()
                    .shortname,
            ))
        });

    let token = jwt::JWTData {
        sub: user.username.as_str(),
        iss: issuer.as_str(),
        aud: client.shortname.as_str(),
        iat: iat.duration_since(std::time::UNIX_EPOCH).unwrap().as_secs(),
        exp: exp.duration_since(std::time::UNIX_EPOCH).unwrap().as_secs(),
        extras: [(
            "roles",
            serde_json::Value::Array(resulting_roles.collect::<Result<Vec<_>, UIDCError>>()?),
        )]
        .into(),
    };

    let key = qi
        .get()
        .by(schema::Key::Realm, &realm.id())
        .one()?
        .ok_or(TokenError::InternalError("no signing key for realm"))?;
    let kpair = ring::signature::Ed25519KeyPair::from_pkcs8(key.keydata.as_slice())
        .map_err(Into::<TokenError>::into)?;

    Ok(jwt::JWT::sign(&kpair, token).into_string())
}

pub fn generate_refresh_token<'a>(
    config: &config::Config,
    qi: &microrm::QueryInterface,
    realm: schema::RealmID,
    client: schema::ClientID,
    user: schema::UserID,
    scopes: impl Iterator<Item = &'a str>,
) -> Result<String, UIDCError> {
    todo!()
}
