use crate::{config::Config, jwt, schema, token, UIDCError};
use microrm::prelude::*;
use ring::signature::KeyPair;

pub fn create_auth_token(
    qi: &microrm::QueryInterface,
    config: &Config,
    realm_id: schema::RealmID,
    client: &str,
    username: &str,
    scopes: &str,
) -> Result<String, UIDCError> {
    token::generate_auth_token(
        config,
        qi,
        realm_id,
        qi.get()
            .only_ids()
            .by(schema::Client::Realm, &realm_id)
            .by(schema::Client::Shortname, client)
            .one_id()?
            .ok_or(UIDCError::Abort("no such client"))?,
        qi.get()
            .only_ids()
            .by(schema::User::Realm, &realm_id)
            .by(schema::User::Username, username)
            .one_id()?
            .ok_or(UIDCError::Abort("no such user"))?,
        scopes.split_whitespace(),
    )
}

pub fn create_refresh_token(
    qi: &microrm::QueryInterface,
    config: &Config,
    realm_id: schema::RealmID,
    client: &str,
    username: &str,
    scopes: &str,
) -> Result<String, UIDCError> {
    token::generate_refresh_token(
        config,
        qi,
        realm_id,
        qi.get()
            .only_ids()
            .by(schema::Client::Realm, &realm_id)
            .by(schema::Client::Shortname, client)
            .one_id()?
            .ok_or(UIDCError::Abort("no such client"))?,
        qi.get()
            .only_ids()
            .by(schema::User::Realm, &realm_id)
            .by(schema::User::Username, username)
            .one_id()?
            .ok_or(UIDCError::Abort("no such user"))?,
        scopes.split_whitespace(),
    )
}

pub fn inspect_token(
    qi: &microrm::QueryInterface,
    config: &Config,
    realm_id: schema::RealmID,
    token: Option<&str>,
) -> Result<(), UIDCError> {
    let key = qi
        .get()
        .by(schema::Key::Realm, &realm_id)
        .one()?
        .ok_or(UIDCError::Abort("no key for realm"))?;

    let kpair = ring::signature::Ed25519KeyPair::from_pkcs8(key.keydata.as_slice())
        .map_err(|_| UIDCError::Abort("could not load key"))?;

    let pubkey = ring::signature::UnparsedPublicKey::new(
        &ring::signature::ED25519,
        kpair.public_key().as_ref(),
    );

    let token = match token {
        Some(token) => token.to_string(),
        None => rpassword::prompt_password("Enter token: ").unwrap(),
    };

    let jwt = jwt::JWT::verify(&pubkey, token.as_str());
    if jwt.is_none() {
        println!("Signature validation against realm key failed!");
    } else if let Some(claims) = jwt.as_ref().and_then(jwt::JWT::claims) {
        println!("Base claims:");
        println!(" - issuer      : {}", claims.iss);
        println!(" - audience    : {}", claims.aud);
        println!(" - subject     : {}", claims.sub);
        println!(
            " - issued at   : {} [{}]",
            claims.iat,
            chrono::DateTime::<chrono::Utc>::from_timestamp(claims.iat as i64, 0)
                .unwrap()
                .to_rfc2822()
        );
        println!(
            " - expires at  : {} [{}]",
            claims.exp,
            chrono::DateTime::<chrono::Utc>::from_timestamp(claims.exp as i64, 0)
                .unwrap()
                .to_rfc2822()
        );
        for claim in claims.extras {
            println!(" - {:12}: {}", claim.0, claim.1);
        }
    } else {
        println!("Claim parsing failed!");
    }
    Ok(())
}
